/***************************************************************************
    name                 : wmartagu.c
    version              : 0.3.1
    description          : Display or modify a WMA file's title and author
                           tags, includes UTF-8 support.
    compile with         :
    cc -Wall -g `pkg-config --cflags --libs glib-2.0` wmartagu.c -o wmartagu
    usage                : wmartagu [-a "Author"] [-t "Title"] "File.wma"
                             -------------------
    begin                : 29 March, 2009
    copyright            : (C) 2009 by Kamran Riaz Khan
    email                : krkhan@inspirated.com
    website              : http://inspirated.com/
 **************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 **************************************************************************/

#include <glib.h>
#include <string.h>
#include <locale.h>

#undef DEBUG
#define TFLAG 1
#define AFLAG 1 << 1
#define ASFCDOGUID \
	"75B22633-668E-11CF-A6D9-00AA0062CE6C"
/*
 * ASF Content Description Object GUID, refer ASF specification at:
 * http://www.microsoft.com/windows/windowsmedia/forpros/format/asfspec.aspx
 */

static gchar *catitle;
static gchar *caauthor;
static gchar **cafiles;
static GOptionEntry entries[] =
{
	{ "title", 't', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_STRING, &catitle, "New title", NULL },
	{ "author", 'a', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_STRING, &caauthor, "New author", NULL},
	{ G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY, &cafiles, NULL, NULL },
	{ NULL }
};

typedef guchar BYTE[1];
typedef guchar WCHAR[2];
typedef guchar WORD[2];
typedef guchar DWORD[4];
typedef guchar QWORD[8];
typedef guchar GUID[16];
typedef gchar GUIDSTR[37];

typedef struct {
	GUIDSTR guid_str;
	GUID guid;
	gsize size;
} AsfObject;

typedef struct {
	gchar *title;
	gchar *author;
	gchar *copyright;
	gchar *desc;
	gchar *rating;
} ContentInfo;

typedef struct {
	gchar *buffer;
	gsize len;
	gsize cur;
} FileBuffer;

void format_guid(GUID guid, GUIDSTR guid_str)
{
	g_snprintf(guid_str, 37,
			"%02X%02X%02X%02X-%02X%02X-%02X%02X-%02X%02X-%02X%02X%02X%02X%02X%02X",
			guid[3], guid[2], guid[1], guid[0],
			guid[5], guid[4],
			guid[7], guid[6],
			guid[8], guid[9],
			guid[10], guid[11], guid[12], guid[13], guid[14], guid[15]);
}

void read_buffer(guchar *buffer, FileBuffer *fb, gsize len)
{
	memcpy(buffer, fb->buffer + fb->cur, len);
	fb->cur += len;
}

void write_buffer(FileBuffer *fb, guchar *buffer, gsize len)
{
	memcpy(fb->buffer + fb->cur, buffer, len);
	fb->cur += len;
}

glong le_to_long(guchar *bytes, gsize size)
{
	glong value;
	int i;

	value = 0;
	for(i = size - 1; i >= 0; i--) {
		value <<= 8;
		value |= bytes[i];
	}

	return value;
}

void long_to_le(guchar *bytes, glong value, gsize size)
{
	glong tmp;
	int i;

	for(i = 0; i < size; i++) {
		tmp = value;
		tmp <<= (sizeof(glong) - 1 - i) * 8;
		tmp >>= (sizeof(glong) - 1 - i) * 8;
		tmp >>= i * 8;
		bytes[i] = tmp;
	}
}

glong word_to_long(WORD word)
{
	return le_to_long(word, 2);
}

glong dword_to_long(DWORD dword)
{
	return le_to_long(dword, 4);
}

glong qword_to_long(QWORD qword)
{
	return le_to_long(qword, 8);
}

void long_to_word(WORD word, glong value)
{
	long_to_le(word, value, 2);
}

void long_to_dword(DWORD dword, glong value)
{
	long_to_le(dword, value, 4);
}

void long_to_qword(QWORD qword, glong value)
{
	long_to_le(qword, value, 8);
}

gchar *read_utf16le(FileBuffer *fb, gsize len)
{
	gchar *utf16, *utf8;

	if(len) {
		utf16 = g_malloc(len);
		read_buffer((guchar *) utf16, fb, len);
		utf8 = g_convert(utf16, len, "UTF-8", "UTF-16LE", NULL, NULL, NULL);
		g_free(utf16);
	} else {
		utf8 = g_malloc(1);
		utf8[0] = '\0';
	}

	return utf8;
}

guchar *utf8_to_utf16le(gsize *len, gchar *utf8)
{
	gsize llen, in;
	guchar *tmp, *utf16;

	tmp = (guchar *) g_convert(utf8, -1, "UTF-16LE", "UTF-8", &in, &llen, NULL);
	if(llen) {
		llen += 2;
		utf16 = g_malloc(llen);
		memcpy(utf16, tmp, llen);
		utf16[llen - 1] = utf16[llen - 2] = '\0';
	}

	if(len != NULL) {
		*len = llen;
	}

	g_free(tmp);

	return utf16;
}

void print_asf_object(AsfObject obj)
{
#ifdef DEBUG
	g_print("OBJ: %s [%ld]\n", obj.guid_str, obj.size);
#endif
}

AsfObject read_asf_object(FileBuffer *fb)
{
	QWORD qword;
	AsfObject obj;

	read_buffer(obj.guid, fb, 16);
	format_guid(obj.guid, obj.guid_str);

	read_buffer(qword, fb, 8);
	obj.size = qword_to_long(qword);

	return obj;
}

void print_content_info(ContentInfo ci)
{
	g_print("Title: %s\n", ci.title);
	g_print("Author: %s\n", ci.author);
	g_print("Copyright: %s\n", ci.copyright);
	g_print("Description: %s\n", ci.desc);
	g_print("Rating: %s\n", ci.rating);
}

ContentInfo read_content_info(FileBuffer *fb)
{
	ContentInfo ci;
	WORD word;
	gsize titlelen, authorlen, copyrightlen, desclen, ratinglen;

	read_buffer(word, fb, 2);
	titlelen = word_to_long(word);

	read_buffer(word, fb, 2);
	authorlen = word_to_long(word);

	read_buffer(word, fb, 2);
	copyrightlen = word_to_long(word);

	read_buffer(word, fb, 2);
	desclen = word_to_long(word);

	read_buffer(word, fb, 2);
	ratinglen = word_to_long(word);

	ci.title = read_utf16le(fb, titlelen);
	ci.author = read_utf16le(fb, authorlen);
	ci.copyright = read_utf16le(fb, copyrightlen);
	ci.desc = read_utf16le(fb, desclen);
	ci.rating = read_utf16le(fb, ratinglen);

	return ci;
}

FileBuffer replace_content_info(const FileBuffer *fb, ContentInfo ci, AsfObject obj)
{
	WORD word;
	DWORD dword;
	QWORD qword;
	FileBuffer newfb;
	gsize titlelen, authorlen, copyrightlen, desclen, ratinglen, hsize, oldsize;
	guchar *u16title, *u16author, *u16copyright, *u16desc, *u16rating;

	oldsize = obj.size;

	newfb.cur = newfb.len = fb->cur - oldsize + 16;
	newfb.buffer = g_malloc(newfb.len);
	memcpy(newfb.buffer, fb->buffer, newfb.len);

	u16title = utf8_to_utf16le(&titlelen, ci.title);
	u16author = utf8_to_utf16le(&authorlen, ci.author);
	u16copyright = utf8_to_utf16le(&copyrightlen, ci.copyright);
	u16desc = utf8_to_utf16le(&desclen, ci.desc);
	u16rating = utf8_to_utf16le(&ratinglen, ci.rating);

	obj.size = 34 + titlelen + authorlen + copyrightlen + desclen + ratinglen;
	newfb.len += obj.size - 16;
	newfb.buffer = g_realloc(newfb.buffer, newfb.len);

	long_to_qword(qword, obj.size);
	write_buffer(&newfb, qword, 8);

	long_to_qword(word, titlelen);
	write_buffer(&newfb, word, 2);

	long_to_qword(word, authorlen);
	write_buffer(&newfb, word, 2);

	long_to_qword(word, copyrightlen);
	write_buffer(&newfb, word, 2);

	long_to_qword(word, desclen);
	write_buffer(&newfb, word, 2);

	long_to_qword(word, ratinglen);
	write_buffer(&newfb, word, 2);

	write_buffer(&newfb, u16title, titlelen);
	write_buffer(&newfb, u16author, authorlen);
	write_buffer(&newfb, u16copyright, copyrightlen);
	write_buffer(&newfb, u16desc, desclen);
	write_buffer(&newfb, u16rating, ratinglen);

	newfb.len += fb->len - fb->cur;
	newfb.buffer = g_realloc(newfb.buffer, newfb.len);
	memcpy(newfb.buffer + newfb.cur, fb->buffer + fb->cur, fb->len - fb->cur);

	newfb.cur = 16;
	read_buffer(dword, &newfb, 4);
	hsize = dword_to_long(dword);
	if(obj.size < oldsize) {
		hsize -= oldsize - obj.size;
	} else {
		hsize += obj.size - oldsize;
	}
	newfb.cur = 16;
	long_to_dword(dword, hsize);
	write_buffer(&newfb, dword, 4);
	newfb.len -= 4;

	print_content_info(ci);

	if(ratinglen) {
		g_free(u16rating);
	}
	if(desclen) {
		g_free(u16desc);
	}
	if(copyrightlen) {
		g_free(u16copyright);
	}
	if(authorlen) {
		g_free(u16author);
	}
	if(titlelen) {
		g_free(u16title);
	}

	return newfb;
}

void free_content_info(ContentInfo *ci)
{
	g_free(ci->title);
	g_free(ci->author);
	g_free(ci->copyright);
	g_free(ci->desc);
	g_free(ci->rating);
}

gint main(gint argc, gchar *argv[])
{
	GOptionContext *context;
	AsfObject obj;
	FileBuffer fb, mb;
	ContentInfo ci;
	DWORD dword;
	glong objcount;
	gchar *file, *help;
	gint i;

	setlocale(LC_ALL, "");

	context = g_option_context_new("File.wma");
	g_option_context_set_summary(context, "Read/modify basic metadata tags in WMA files");
	g_option_context_add_main_entries(context, entries, NULL);

	g_set_prgname(argv[0]);

	if(argc < 2) {
		help = g_option_context_get_help(context, FALSE, NULL);
		g_printerr("%s\n", help);
		g_free(help);
		g_option_context_free(context);
		return 1;
	}

	if(!g_option_context_parse(context, &argc, &argv, NULL)) {
		help = g_option_context_get_help(context, FALSE, NULL);
		g_printerr("Option parsing failed\n");
		g_printerr("%s\n", help);
		g_free(help);
		g_option_context_free(context);
		return 1;
	}

	g_option_context_free(context);

	if(cafiles) {
		for(i = 0; cafiles[i]; i++);
		file = cafiles[i - 1];
	} else {
		g_printerr("No filename specified\n");
		return 2;
	}

	if(!g_file_get_contents(file, &fb.buffer, &fb.len, NULL)) {
		g_printerr("File error\n");
		return 2;
	}
	fb.cur = 0;

	obj = read_asf_object(&fb);
	print_asf_object(obj);

	read_buffer(dword, &fb, 4);
	objcount = dword_to_long(dword);

	if(objcount < 1) {
		g_printerr("Invalid header count\n");
		return 4;
	}

	/* Skip reserved bytes */
	fb.cur += 2;

	for(i = 0; i < objcount; i++) {
		obj = read_asf_object(&fb);
		print_asf_object(obj);

		if(!g_strcmp0(obj.guid_str, ASFCDOGUID)) {
			ci = read_content_info(&fb);

			if(catitle) {
				g_free(ci.title);
				ci.title = catitle;
			}
			if(caauthor) {
				g_free(ci.author);
				ci.author = g_strdup(caauthor);
			}

			if(catitle || caauthor) {
				mb = replace_content_info(&fb, ci, obj);
				if(!g_file_set_contents(file, mb.buffer, mb.len, NULL)) {
					g_printerr("File error");
					return 2;
				}
				g_free(mb.buffer);

				g_print("Tags updated successfully.\n");
			} else {
				print_content_info(ci);
			}

			free_content_info(&ci);
		} else {
			fb.cur += obj.size - 24;
		}
	}

	g_free(fb.buffer);

	return 0;
}

